// Converts an instance of Instant Insanity to HCP.
//
// The instance should be in the following format. The first line has the number of cubes/colours, then each 
// subsequent line has the colour for each of the six faces of a cube. The order of colours corresponds to
// the following flat layout of a cube:
// 
//      ___
//     | 1 |
//  ___|___|___
// | 2 | 3 | 4 |
// |___|___|___|
//     | 5 |
//     |___|
//     | 6 |
//     |___|
//
// So the following input:
// 6
// 5 1 4 6 3 2
// 5 3 4 6 1 2
// 3 5 4 6 2 1
// 3 2 4 6 5 1
// 2 3 4 6 1 5
// 1 5 4 6 2 3
//
// Corresponds to the following six cubes:
//
//      ___              ___              ___              ___              ___              ___
//     | 5 |            | 5 |            | 3 |            | 3 |            | 2 |            | 1 |    
//  ___|___|___      ___|___|___      ___|___|___      ___|___|___      ___|___|___      ___|___|___
// | 1 | 4 | 6 |    | 3 | 4 | 6 |    | 5 | 4 | 6 |    | 2 | 4 | 6 |    | 3 | 4 | 6 |    | 5 | 4 | 6 |
// |___|___|___|    |___|___|___|    |___|___|___|    |___|___|___|    |___|___|___|    |___|___|___|
//     | 3 |            | 1 |            | 2 |            | 5 |            | 1 |            | 2 |
//     |___|            |___|            |___|            |___|            |___|            |___|
//     | 2 |            | 2 |            | 1 |            | 1 |            | 5 |            | 3 |
//     |___|            |___|            |___|            |___|            |___|            |___|
//
   //
//
// Written by Michael Haythorpe


import java.io.*;
   import java.util.StringTokenizer;
   import java.util.Arrays;

   public class II2HCP
   {
      public static void main(String [] args)
      {
         if(args.length < 1 || args.length > 2)
         {
            System.out.println("II2HCP should be run using the following input:");
            System.out.println();
            System.out.println("II2HCP.exe <input_file.ii> <output_file.hcp>");
            System.out.println();
            System.out.println("If no output filename is specified, COL2HCP creates input_file.hcp.");
            return;
         }
      
         String inputFilename = "";
         String outputFilename = "";
      
         if(args.length == 1)
         {
            inputFilename = args[0];
            outputFilename = inputFilename;
            if(outputFilename.substring(outputFilename.length()-3).equals(".ii"))
               outputFilename = outputFilename.substring(0,outputFilename.length()-3);
            outputFilename += ".hcp";
         }
      
         if(args.length == 2)
         {
            inputFilename = args[0];
            outputFilename = args[1];
         }
      
         File inputFile = new File(inputFilename);
         File outputFile = new File(outputFilename);
      
         int numberCubes;
         int [][]faces;
      
         BufferedReader br;
         try
         {
            br = new BufferedReader(new FileReader(inputFile));
         }
            catch(Exception e)
            {
               System.out.println("ERROR: " + inputFilename + " can not be opened!");
               return;
            }
      
         String line = "";
         try
         {
            line = br.readLine();
            numberCubes = Integer.parseInt(line);
         }
            catch(Exception e)
            {
               System.out.println("ERROR: The first line of " + inputFilename + " should be an integer! Instead it is: " + line);
               return;
            }
      
         faces = new int[numberCubes][6];
      
         try
         {
            for(int i=0; i<numberCubes; i++)
            {
               line = br.readLine();
               StringTokenizer tokens = new StringTokenizer(line);
               for(int j=0; j<6; j++)
               {
                  faces[i][j] = Integer.parseInt(tokens.nextToken());
                  if(faces[i][j] > numberCubes || faces[i][j] <= 0)
                  {
                     System.out.println("ERROR: Each line of " + inputFilename + " should contain integers between 1 and " + numberCubes + " for the six faces. One line contained: " + line);
                     return;
                  }   
               }
            }
         
         }
            catch(Exception e)
            {
               System.out.println("ERROR: Each line of " + inputFilename + " should contain integers between 1 and " + numberCubes + " for the six faces. One line contained: " + line);
               return;
            }
      
         try
         {
            br.close();
         }
            catch(Exception e)
            {
               System.out.println("ERROR: Couldn't close " + inputFilename);
               return;
            }
      
      // Now to start constructing graph
      
         int N = 66*numberCubes + 4;
         int [][]arcs = new int[N][3*N];
         int []degrees = new int[N];
      
         arcs[0][0] = 1;
         degrees[0]++;
      
      // Build bottom network
      
         for(int i=0; i<numberCubes; i++)
            for(int j=0; j<numberCubes; j++)
            {
               arcs[2*(i+1)][degrees[2*(i+1)]++] = 2*j+1;
               arcs[2*(i+1) + 2*numberCubes + 1][degrees[2*(i+1) + 2*numberCubes + 1]++] = 2*j+1 + 2*numberCubes + 1;
            }
      
      // Connect bottom network to in-between vertices
      
         for(int i=0; i<numberCubes; i++)
         {
            arcs[2*(i+1)][degrees[2*(i+1)]++] = 2*numberCubes + 1;
            arcs[2*(i+1) + 2*numberCubes + 1][degrees[2*(i+1) + 2*numberCubes + 1]++] = 2*numberCubes + 1 + 2*numberCubes + 1;
         }
      
         arcs[2*numberCubes + 1][degrees[2*numberCubes + 1]++] = 2*numberCubes + 2;
         arcs[4*numberCubes + 2][degrees[4*numberCubes + 2]++] = 4*numberCubes + 3;
      
      // Build first upper network
      
         int start = 4*numberCubes + 3;
      
         for(int i=0; i<numberCubes; i++)
         {
            for(int j=0; j<3; j++)
            {
               arcs[start+28*i][degrees[start+28*i]++] = start+28*i+9*j+1; // Connect starting vertex to each face vertex
               arcs[start+28*i+9*j+1][degrees[start+28*i+9*j+1]++] = start+28*i+9*j+2;   // Connect each face vertex to adjacent hexagon
            
            	// Now build each hexagon
               
               arcs[start+28*i+9*j+2][degrees[start+28*i+9*j+2]++] = start+28*i+9*j+3;
               arcs[start+28*i+9*j+2][degrees[start+28*i+9*j+2]++] = start+28*i+9*j+9;
               arcs[start+28*i+9*j+3][degrees[start+28*i+9*j+3]++] = start+28*i+9*j+2;
               arcs[start+28*i+9*j+3][degrees[start+28*i+9*j+3]++] = start+28*i+9*j+4;
               arcs[start+28*i+9*j+4][degrees[start+28*i+9*j+4]++] = start+28*i+9*j+3;
               arcs[start+28*i+9*j+4][degrees[start+28*i+9*j+4]++] = start+28*i+9*j+5;
               arcs[start+28*i+9*j+5][degrees[start+28*i+9*j+5]++] = start+28*i+9*j+4;
               arcs[start+28*i+9*j+5][degrees[start+28*i+9*j+5]++] = start+28*i+9*j+6;
               arcs[start+28*i+9*j+6][degrees[start+28*i+9*j+6]++] = start+28*i+9*j+5;
               arcs[start+28*i+9*j+6][degrees[start+28*i+9*j+6]++] = start+28*i+9*j+7;
               arcs[start+28*i+9*j+7][degrees[start+28*i+9*j+7]++] = start+28*i+9*j+6;
               arcs[start+28*i+9*j+7][degrees[start+28*i+9*j+7]++] = start+28*i+9*j+8;
               arcs[start+28*i+9*j+8][degrees[start+28*i+9*j+8]++] = start+28*i+9*j+7;
               arcs[start+28*i+9*j+8][degrees[start+28*i+9*j+8]++] = start+28*i+9*j+9;
               arcs[start+28*i+9*j+9][degrees[start+28*i+9*j+9]++] = start+28*i+9*j+2;
               arcs[start+28*i+9*j+9][degrees[start+28*i+9*j+9]++] = start+28*i+9*j+3;
               arcs[start+28*i+9*j+9][degrees[start+28*i+9*j+9]++] = start+28*i+9*j+8;
            }
             
             // Connect the three hexagons to the subsequent face vertices
             
            arcs[start+28*i+9][degrees[start+28*i+9]++] = start+28*i+10;
            arcs[start+28*i+18][degrees[start+28*i+18]++] = start+28*i+19;
            arcs[start+28*i+27][degrees[start+28*i+27]++] = start+28*i+1;	
         }
         
        
      	
      // Build second upper network
      
         start = 32*numberCubes + 3;
      
         for(int i=0; i<numberCubes; i++)
         {
            for(int j=0; j<3; j++)
            {
            
               arcs[start+34*i][degrees[start+34*i]++] = start+34*i+11*j+1; // Connect starting vertex to each face diamond
            
            // Build each face diamond
            
               arcs[start+34*i+11*j+1][degrees[start+34*i+11*j+1]++] = start+34*i+11*j+2;
               arcs[start+34*i+11*j+2][degrees[start+34*i+11*j+2]++] = start+34*i+11*j+1;
               arcs[start+34*i+11*j+2][degrees[start+34*i+11*j+2]++] = start+34*i+11*j+3;
               arcs[start+34*i+11*j+3][degrees[start+34*i+11*j+3]++] = start+34*i+11*j+2;
               
            // Connect each face diamond to alternative two hexagons
            
               for(int k=0; k<3; k++)
                  if(k != j)
                     arcs[start+34*i+11*j+3][degrees[start+34*i+11*j+3]++] = start+34*i+11*k+4;
                     
            // Now build each hexagon
            
               arcs[start+34*i+11*j+4][degrees[start+34*i+11*j+4]++] = start+34*i+11*j+5;
               arcs[start+34*i+11*j+4][degrees[start+34*i+11*j+4]++] = start+34*i+11*j+11;
               arcs[start+34*i+11*j+5][degrees[start+34*i+11*j+5]++] = start+34*i+11*j+4;
               arcs[start+34*i+11*j+5][degrees[start+34*i+11*j+5]++] = start+34*i+11*j+6;
               arcs[start+34*i+11*j+6][degrees[start+34*i+11*j+6]++] = start+34*i+11*j+5;
               arcs[start+34*i+11*j+6][degrees[start+34*i+11*j+6]++] = start+34*i+11*j+7;
               arcs[start+34*i+11*j+7][degrees[start+34*i+11*j+7]++] = start+34*i+11*j+6;
               arcs[start+34*i+11*j+7][degrees[start+34*i+11*j+7]++] = start+34*i+11*j+8;
               arcs[start+34*i+11*j+8][degrees[start+34*i+11*j+8]++] = start+34*i+11*j+7;
               arcs[start+34*i+11*j+8][degrees[start+34*i+11*j+8]++] = start+34*i+11*j+9;
               arcs[start+34*i+11*j+9][degrees[start+34*i+11*j+9]++] = start+34*i+11*j+8;
               arcs[start+34*i+11*j+9][degrees[start+34*i+11*j+9]++] = start+34*i+11*j+10;
               arcs[start+34*i+11*j+10][degrees[start+34*i+11*j+10]++] = start+34*i+11*j+9;
               arcs[start+34*i+11*j+10][degrees[start+34*i+11*j+10]++] = start+34*i+11*j+11;
               arcs[start+34*i+11*j+11][degrees[start+34*i+11*j+11]++] = start+34*i+11*j+4;
               arcs[start+34*i+11*j+11][degrees[start+34*i+11*j+11]++] = start+34*i+11*j+5;
               arcs[start+34*i+11*j+11][degrees[start+34*i+11*j+11]++] = start+34*i+11*j+10;
            
                 
            // Connect each hexagon to alternative two diamonds
            
               for(int k=0; k<3; k++)
                  if(k != j)
                     arcs[start+34*i+11*j+11][degrees[start+34*i+11*j+11]++] = start+34*i+11*k+1;
            
            // Connect each hexagon to starting vertex of next structure
            
               arcs[start+34*i+11*j+11][degrees[start+34*i+11*j+11]++] = start+34*i+34;
            }
         	
         }
      
         
      	// Close final vertex
      	
         arcs[N-1][0] = 0;
         degrees[N-1]++;
         
      	// Joins between two upper networks
      	
         for(int i=0; i<numberCubes; i++)
            for(int j=0; j<3; j++)
            {
               arcs[4*numberCubes+3+28*i+9*j+1][degrees[4*numberCubes+3+28*i+9*j+1]++] = start+34*i+11*j+3;
               arcs[start+34*i+11*j+1][degrees[start+34*i+11*j+1]++] = 4*numberCubes+3+28*i+28;
            }
            
         // Finally, introduce appropriate edges for each cube
      	
         int orientations[][] = new int[3][2];
         orientations[0][0] = 0;
         orientations[0][1] = 4;
         orientations[1][0] = 1;
         orientations[1][1] = 3;
         orientations[2][0] = 2;
         orientations[2][1] = 5;
      	
         for(int i=0; i<numberCubes; i++)
            for(int j=0; j<3; j++)
            {
            
            // For first bottom to first upper
            
               arcs[faces[i][orientations[j][0]]*2-1][degrees[faces[i][orientations[j][0]]*2-1]++] = 4*numberCubes+3+28*i+9*j+4;  // From bottom to upper
               arcs[4*numberCubes+3+28*i+9*j+2][degrees[4*numberCubes+3+28*i+9*j+2]++] = faces[i][orientations[j][1]]*2; // From upper to bottom
               if(faces[i][orientations[j][0]] != faces[i][orientations[j][1]])  // Repeat unless the colours are the same
               {
                  arcs[faces[i][orientations[j][1]]*2-1][degrees[faces[i][orientations[j][1]]*2-1]++] = 4*numberCubes+3+28*i+9*j+6;
                  arcs[4*numberCubes+3+28*i+9*j+7][degrees[4*numberCubes+3+28*i+9*j+7]++] = faces[i][orientations[j][0]]*2;
               }
            
             // For second bottom to second upper
            
               arcs[2*numberCubes+1+faces[i][orientations[j][0]]*2-1][degrees[2*numberCubes+1+faces[i][orientations[j][0]]*2-1]++] = start+34*i+11*j+6;  // From bottom to upper
               arcs[start+34*i+11*j+4][degrees[start+34*i+11*j+4]++] = 2*numberCubes+1+faces[i][orientations[j][1]]*2; // From upper to bottom
               if(faces[i][orientations[j][0]] != faces[i][orientations[j][1]])  // Repeat unless the colours are the same
               {
                  arcs[2*numberCubes+1+faces[i][orientations[j][1]]*2-1][degrees[2*numberCubes+1+faces[i][orientations[j][1]]*2-1]++] = start+34*i+11*j+8;
                  arcs[start+34*i+11*j+9][degrees[start+34*i+11*j+9]++] = 2*numberCubes+1+faces[i][orientations[j][0]]*2;
               }
            
            }
            
      		// Now to convert to undirected HCP
      		
         int [][]undirectedEdges = new int[3*N][3*N];
         int []undirectedDegrees = new int[3*N];
         for(int i=0; i<N; i++)
         {
            undirectedEdges[3*i][undirectedDegrees[3*i]++] = 3*i+1;
            undirectedEdges[3*i+1][undirectedDegrees[3*i+1]++] = 3*i;
            undirectedEdges[3*i+1][undirectedDegrees[3*i+1]++] = 3*i+2;
            undirectedEdges[3*i+2][undirectedDegrees[3*i+2]++] = 3*i+1;
         }
         
         for(int i=0; i<N; i++)
            for(int j=0; j<degrees[i]; j++)
            {
               undirectedEdges[3*i+2][undirectedDegrees[3*i+2]++] = 3*arcs[i][j];
               undirectedEdges[3*arcs[i][j]][undirectedDegrees[3*arcs[i][j]]++] = 3*i+2;
            }
      		
         for(int i=0; i<3*N; i++)
         {
         // Now to sort each row
            int []tempRow = new int[undirectedDegrees[i]];
            for(int j=0; j<undirectedDegrees[i]; j++)
               tempRow[j] = undirectedEdges[i][j];
         
            Arrays.sort(tempRow);
         
            for(int j=0; j<undirectedDegrees[i]; j++)
               undirectedEdges[i][j] = tempRow[j];
         }
      		
      		
      		
         try
         {
            BufferedWriter bw = new BufferedWriter(new FileWriter(outputFilename));
         	
            bw.write("NAME : Adapted from II file " + inputFilename);
            bw.newLine();
            bw.write("COMMENT : ");
            bw.newLine();
            bw.write("TYPE : HCP");
            bw.newLine();
            bw.write("DIMENSION : " + 3*N);
            bw.newLine();
            bw.write("EDGE_DATA_FORMAT : EDGE_LIST");
            bw.newLine();
            bw.write("EDGE_DATA_SECTION");
            bw.newLine();
         
            for(int i=0; i<3*N; i++)
               for(int j=0; j<undirectedDegrees[i]; j++)
               {
                  bw.write((i+1) + " " + (undirectedEdges[i][j]+1));
                  bw.newLine();
               }
            
            bw.write("-1");
            bw.newLine();
            bw.write("EOF");
            bw.close();
         }
            catch(Exception e)
            {
               System.out.println("ERROR: Unable to write to file " + outputFilename);
               return;
            }
      }   
   }